#!/usr/bin/env perl
## 
## Copyright (C) 2011-2012 The Regents of the University of California.
## 
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
## 
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
## 
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
## 

# Utilities for MIDAR wrapper scripts

use strict;

package MidarWrapper;
use Exporter;
our (@ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
@ISA = qw(Exporter);
@EXPORT = qw($autostep $autostep_stop $bogons $run_id $distrib
    $mper_port $mper_pps
    $est_round_duration $est_overlap $disc_overlap $elim_overlap $cor_overlap
    $est_rounds $elim_rounds $cor_rounds $disc_responsive $clockoff $n_threads
    $elim_concurrency $cor_concurrency
    $nice $keep_small_untested $bindir $utildir
    $step $num $starttime @children $datadir $andir $stepdir
    $usage_options $usage_steps $usage_extra
    @options @monitors @targetfiles $exclude_monitors
    $STATS
    init start_step next_step parallel_execute reap analysis_plots
    check_run_num newdir runprog check_concurrency choose2
    midar_step);
@EXPORT_OK = qw();

#####################################################

use strict;
use Getopt::Long qw(GetOptionsFromString);
use List::Util qw(shuffle);
use FindBin qw($Bin $Script);
use Cwd;
use Cwd 'abs_path';
use File::Glob ':glob';
use File::Basename;
use POSIX qw(strftime floor ceil);
use autodie;

my $cvsid = '$Id: MidarWrapper.pm,v 1.18 2012/07/11 21:40:16 kkeys Exp $';


#####################################################
# configuration

our $autostep = 0;
our $autostep_stop;
our $bogons = undef;
our $run_id = undef;

our $distrib = 0;
our @targetfiles;
our $mper_port = 8746;
our $mper_pps = 100;
our $est_round_duration = 10;
our $est_overlap = 10;
our $disc_overlap = 5;
our $elim_overlap = 5;
our $cor_overlap = 5;
our $est_rounds = 30;
#our $disc_rounds = 30; # not used
our $elim_rounds = 10;
our $cor_rounds = 10;
our $disc_responsive = 23; # assumes disc_min_rounds=30
our $clockoff = 100; # irrelevant in single-monitor run
our $n_threads = 1;
our $elim_concurrency;
our $cor_concurrency;

our $nice = ""; # nice="nice -n1"
our $keep_small_untested = 1;
our $simulate;

our $bindir="$Bin";
our $utildir="$Bin/../utils";

our $step;
our $num;
our $starttime;
our @children;
our $datadir;
our $andir;
our $stepdir;

our @monitors;
our $exclude_monitors;
our $MIDAR_MONS;

our $STATS;

our $dataparent = '.';
our $anparent = '.';

our $monitortxt;
our $monitoryaml;
our $monitoryaml_default = "monitors.yaml";

###########################################################################
# options

our @options = (
    "distributed!" => \$distrib,
    "autostep!" => \$autostep,
    "stop=s" => \$autostep_stop,
    "targets=s" => \@targetfiles,
    "datadir=s" => \$dataparent,
    "andir=s" => \$anparent,
    "dir=s" => sub {$dataparent=$_[1]; $anparent=$_[1];},
    "run-id=s" => \$run_id,
    "mper-port=i" => \$mper_port,
    "threads=i" => \$n_threads,
    "mper-pps=i" => \$mper_pps,
    "est-duration=i" => \$est_round_duration,
    "est-rounds=i" => \$est_rounds,
    "elim-rounds=i" => \$elim_rounds,
    "cor-rounds=i" => \$cor_rounds,
    "est-overlap=i" => \$est_overlap,
    "disc-overlap=i" => \$disc_overlap,
    "elim-overlap=i" => \$elim_overlap,
    "cor-overlap=i" => \$cor_overlap,
    "monitors=s" => \$monitoryaml,
    "exclude-monitors=s" => \$exclude_monitors,
    "simulate!" => \$simulate,
    );

our $usage_options =
	"    --run-id=<RUN_ID>      label for this experiment, e.g. start date (REQUIRED)\n" .
	"    --[no-]distrib         distributed mode: probe from multiple monitors\n" .
	"    --config=<FILE>        read additional command-line options from <FILE>;\n" .
	"                           lines beginning with \"#\" will be ignored\n" .
	"    --[no-]autostep        automatically proceed to next step\n" .
	"    --stop=<STEP>          autostep up to but not including <STEP>\n" .
	"    --targets=<FILE>       load probe target addresses from <FILE> (repeatable)\n" .
	"    --datadir=<DIR>        write probe results to <DIR>/run-<RUN_ID> [${dataparent}]\n" .
	"    --andir=<DIR>          write analysis results to <DIR>/run-<RUN_ID> [${anparent}]\n" .
	"    --dir=<DIR>            shorthand for --datadir=<DIR> --andir=<DIR>\n" .
	"    --threads=<N>          use up to <N> threads or processes [$n_threads]; should\n" .
	"                           be <= number of cpu cores\n" .
	"    --mper-port=<PORT>     connect to mper on <PORT> [$mper_port]\n" .
	"    --mper-pps=<N>         assume mper sends <N> pkts per second [$mper_pps]\n" .
	"    --est-duration=<N>     duration of Estimation rounds, in seconds [$est_round_duration]\n" .
	"    --est-rounds=<N>       number of probing rounds in Estimation [$est_rounds]\n" .
	"    --elim-rounds=<N>      number of probing rounds in Elimination [$elim_rounds]\n" .
	"    --cor-rounds=<N>       number of probing rounds in Corroboration [$cor_rounds]\n" .
	"    --est-overlap=<N>      samples needed by MBT in Estimation [$est_overlap]\n" .
	"    --disc-overlap=<N>     samples needed by MBT in Discovery [$disc_overlap]\n" .
	"    --elim-overlap=<N>     samples needed by MBT in Elimination [$elim_overlap]\n" .
	"    --cor-overlap=<N>      samples needed by MBT in Corroboration [$cor_overlap]\n" .
	"    --elim-concurrency=<N> subsets to probe in parallel in Elimination\n" .
	"                           [default is based on target list size and mper-pps]\n" .
	"    --cor-concurrency=<N>  subsets to probe in parallel in Corroboration\n" .
	"                           [default is based on target list size and mper-pps]\n" .
	"    --monitors=<FILE>      monitor file (with --distrib) [\$ARKUTIL_SPEC_PATH\n" .
	"                           or $monitoryaml_default]\n" .
	"    --exclude-mon=<MON>    csv list of monitors to exclude (with --distrib)\n" .
	"    --[no-]simulate        remote monitors should only simulate probing\n";

our $usage_steps =
	"    remote-start (distributed mode only)\n" .
	"    est-targets (expects one or more --targets options)\n" .
	"    est-probe\n" .
	"    est-an\n" .
	"    disc-targets\n" .
	"    disc-probe\n" .
	"    disc-an\n" .
	"    elim-targets [<NUM>]\n" .
	"    elim-probe [<NUM>]\n" .
	"    elim-an [<NUM>]\n" .
	"    elim-final\n" .
	"    cor-targets [<NUM>]\n" .
	"    cor-probe [<NUM>]\n" .
	"    cor-an [<NUM>]\n" .
	"    midar-an\n" .
	"    remote-stop (distributed mode only)\n";

our $usage_extra = "";

###########################################################################

sub runprog($$$$) {
    my ($prog, $errtype, $errfile, $cmd) = @_;
    # prog: name of program (for reporting purposes)
    # errtype: 'file', 'tee', or undef
    # errfile: name of error file
    # cmd: command line (not including redirects covered by errtype/errfile)

    print STDERR "# executing: $cmd\n";
    my $progbase = basename($prog);
    my $saveerr;
    if ($errtype eq 'file') {
	open $saveerr, '>&', \*STDERR;
	open STDERR, '>', $errfile; # redirect to file
    } elsif ($errtype eq 'tee') {
	open $saveerr, '>&', \*STDERR;
	open STDERR, '|-', "tee $errfile"; # redirect to tee
	$errfile = "above or $errfile";
    } else {
	$errfile = 'above';
    }
    system($cmd);
    my $status = $?;
    my $errno = $!;
    if ($errtype) {
	open STDERR, '>&', $saveerr; # restore
    }
    if ($status == -1) {
        die "failed to execute $prog: $errno\n";
    } elsif ($status & 127) {
        die "\n$prog died with signal " . ($status & 127) .
	    ".  See $errfile for error messages.\n";
    } elsif (($status >> 8) != 0) {
        die "\n$prog exited with value " . ($status >> 8) .
	    ".  See $errfile for error messages.\n";
    }
    print STDERR "# $prog finished successfully.\n\n";
}

sub check_run_num() {
    if (@ARGV == 0) {
	$num = 1;
    } elsif (@ARGV > 1) {
	usageExit();
    } else {
	$num = shift @ARGV;
	die "Error: num must be 1-9\n" unless ($num =~ /^[1-9]$/);
    }
    print STDERR "# step $step, run number $num\n";
}

# bg_wait_gzfiles() { ... }

# wait_file() { ... }

sub parallel_execute($;$) {
    my ($cmd, $msg) = @_;
    while (@children >= $n_threads) {
	my $pid = wait;
	if ($pid == -1) {
	    print STDERR "error: lost children?\n";
	    @children = ();
	    last;
	}
	@children = grep {$_ != $pid} @children;
    }
    print STDERR $msg if $msg;
    my $pid = fork();
    if ($pid) {
	# parent
	push @children, $pid;
    } elsif ($pid == 0) {
	# child
	local $SIG{__DIE__};
	exec($cmd);
	die "exec $cmd:  $!\n";
    } else {
	die "fork: $!\n";
    }
}

sub reap(;$) {
    my ($msg) = @_;
    print STDERR $msg if defined $msg;
    while (@children > 0) {
	my $pid = wait;
	if ($pid == -1) {
	    print STDERR "error: lost children?\n";
	    @children = ();
	    last;
	}
	@children = grep {$_ != $pid} @children;
    }
    print STDERR "...done.\n";
}

sub analysis_plots($$) {
    my ($topdir, $subdir) = @_;
    my $wd = getcwd;
    my $gnuplot = "gnuplot";
    my $movable = "$topdir/$subdir/*.html";
    chdir("$topdir/$subdir");

    system("$gnuplot -e quit");
    my $status = $?;
    my $errno = $!;
    if ($status == -1) {
        warn "failed to execute $gnuplot: $errno" .
	    ".  Report graphs will not be available.\n";
    } elsif ($status & 127) {
        warn "\n$gnuplot died with signal " . ($status & 127) .
	    ".  Report graphs will not be available.\n";
    } elsif (($status >> 8) != 0) {
        warn "\n$gnuplot exited with value " . ($status >> 8) .
	    ".  Report graphs will not be available.\n";
    } else {
	print STDERR "plotting report graphs...\n";
	my @gpfiles = bsd_glob('*.gp', GLOB_ERR);
	die "glob: $!\n" if (GLOB_ERROR);
	for my $gpfile (@gpfiles) {
	    parallel_execute("$gnuplot $gpfile");
	}
	reap();
	$movable .= " $topdir/$subdir/*.png";
    }

    chdir($wd);
    -d "$topdir/html" || mkdir("$topdir/html");
    -d "$topdir/html/$subdir" || mkdir("$topdir/html/$subdir");
    system("mv $movable $topdir/html/$subdir");
    print "Reports written to $topdir/html/$subdir\n\n";
}

sub newdir($) {
    my ($dir) = @_;
    if (-d $dir) {
	die "Directory $dir already exists.  Remove it, or choose a new directory.\n";
    }
    mkdir $dir;
}

sub check_concurrency($$$) {
    my ($ref, $stage, $n_targets) = @_;
    my $max_concurrency = floor($mper_pps * .6 / 2);
    if (!defined $$ref) {
	$$ref = ceil($n_targets / (3600/.6));
	$$ref = 5 if ($$ref < 5);
	$$ref = $max_concurrency if ($$ref > $max_concurrency);
	print STDERR "Automatically setting ${stage}-concurrency to $$ref.\n"
    } else {
	if ($$ref > $max_concurrency) {
	    die "${stage}-concurrency=$$ref is higher than $max_concurrency, the maximum recommended value for mper-pps=$mper_pps\n";
	}
    }
}

sub ftime($) {
    my ($t) = @_;
    return strftime("%F %T", localtime $t);
}

sub next_step($@) {
    my $next;
    ($next, @ARGV) = @_;
    my $now = time;
    my $runtime = $now - $starttime;
    print $STATS "Step $step finished at ", ftime($now), " in ${runtime}s.\n\n";
    print STDERR "\n####### Step $step finished in ${runtime}s.\n";
    if ($autostep && $next ne $autostep_stop) {
	print STDERR "####### Proceeding to next step: $next\n\n";
	$step = $next;
    } else {
	print STDERR "####### Next step: $next" if defined $next;
	print STDERR "\n\n";
	$autostep = 0;
	$step = undef;
    }
}

sub start_step(;$) {
    ($stepdir) = @_;
    newdir($stepdir) if $stepdir;
    $starttime = time;
    print $STATS "#############################################\n";
    print STDERR "#############################################\n";
    print $STATS "Step $step started at ", ftime($starttime), "\n";
    print STDERR "####### Step $step started at ", ftime($starttime), "\n";
}

sub choose2($) {
    my ($n) = @_;
    return $n * ($n - 1) / 2;
}

sub usageExit() {
    print STDERR "Usage: $Script --run-id=<RUN_ID> [<options>] <step> [<args>]\n";
    print STDERR "options, with default values in []:\n";
    print STDERR $usage_options;
    print STDERR "steps (in order):\n";
    print STDERR $usage_steps;
    print STDERR "arguments:\n" .
	"    <target_file>...     file(s) containing a list of IP addresses to resolve\n" .
	"    <num>                repetition number of a step [1]\n";
    print STDERR $usage_extra;
    exit 1;
}

sub read_config_file() {
    my (undef, $file, undef) = @_;
    open(CONFIG, '<', $file) or die "$file: $!\n";
    GetOptionsFromString(join(' ', grep {!/^#/} <CONFIG>), @options) or usageExit();
    close CONFIG;
}

sub remote_manage($) {
    my ($args) = @_;
    runprog("$bindir/manage-midar",
	'tee', "$datadir/manage-midar.err",
	"$bindir/manage-midar " .
	"--mper-port=$mper_port " .
	"--include=$MIDAR_MONS " .
	"--target-dir=$anparent " .
	"--data-dir=$dataparent " .
	"--run-id=$run_id " .
	(defined $num ? "--run-iteration=$num " : '') .
	"--verbose " .
	$args . " >&2");
}

sub remote_probe($) {
    my ($args) = @_;
    runprog("$bindir/midar-remote-control",
	'tee', "$datadir/midar-remote-control.err",
	"$bindir/midar-remote-control " .
	"--mper-port=$mper_port " .
	"--include=$MIDAR_MONS " .
	"--target-dir=$anparent " .
	"--data-dir=$dataparent " .
	"--run-id=$run_id " .
	(defined $num ? "--run-iteration=$num " : '') .
	$args . " >&2");
}

sub cor_style_probe($$$) {
    my ($type, $rounds, $concurrency) = @_;
    my $targetdir = "$andir/${type}${num}-targets";

    if ($distrib) {
	remote_manage("--preprocess-$type-targets");
	remote_manage("--upload-$type-targets");

	# one-src
	my $n_targets = `cat $targetdir/targets.*one-src | egrep -c '^[^#]'`;
	chomp $n_targets;
	if ($n_targets > 0) {
	    check_concurrency(\$concurrency, $type, $n_targets);
	    my $eta = ($n_targets * .61 / $concurrency) * $rounds;
	    print STDERR "$type-one-src estimated probing time: ${eta}s\n";

	    remote_probe(
		"--start-$type-one-src-prober " .
		"--cor-rounds=$rounds " .
		"--cor-concurrency=$concurrency");
	    remote_manage("--download-$type-one-src-results");
	}

	# multi-src
	my $n_targets = `cat $targetdir/targets.*multi-src | egrep -c '^[^#]'`;
	chomp $n_targets;
	if ($n_targets > 0) {
	    # XXX these calculations are probably wrong
	    check_concurrency(\$concurrency, $type, $n_targets);
	    my $eta = ($n_targets * .61 / $concurrency) * $rounds;
	    print STDERR "$type-multi-src estimated probing time: ${eta}s\n";
	    remote_probe(
		"--start-$type-multi-src-prober " .
		"--cor-rounds=$rounds " .
		"--cor-concurrency=$concurrency");
	    remote_manage("--download-$type-multi-src-results");

	    # restart Marinda and shell-server after multi-src run (not needed
	    # after one-src run) to deal with Ruby memory leak
	    remote_manage("--start-marinda --start-shell-server");
	}

	remote_manage("--delete-$type-all");

    } else {
	my $n_targets = `cat $targetdir/targets.*-src | egrep -c '^[^#]'`;
	chomp $n_targets;
	check_concurrency(\$concurrency, $type, $n_targets);
	my $eta = ($n_targets * .61 / $concurrency) * $rounds;
	print STDERR "Estimated probing time: ${eta}s\n";

	runprog("$bindir/midar-cor-one-src-prober",
	    'file', "$stepdir/local.err",
	    "$bindir/midar-cor-one-src-prober " .
	    "--targets=$andir/${type}${num}-targets/targets.one-src " .
	    "--log=$stepdir/local.log " .
	    "--concurrency=$concurrency " .
	    "--rounds=$rounds " .
	    "--mper-port=$mper_port " .
	    ">$stepdir/local.one-src.out");
    }
}

###########################################################################
# initialize

sub init($) {
    my ($cmdline) = @_;

    $usage_options =~ s/\${run_id}/$run_id/g;

    GetOptions("config=s" => \&read_config_file, @options) or usageExit();
    $autostep = 1 if $autostep_stop;

    if (!@ARGV) {
	print STDERR "missing <step>\n";
	usageExit();
    }
    $step = shift @ARGV;

    if ($run_id !~ /^[-+=_a-zA-Z0-9]+/) {
	die "illegal run_id \"$run_id\"\n";
    }

    $monitoryaml ||= $ENV{'ARKUTIL_SPEC_PATH'} || $monitoryaml_default;
    $monitoryaml = abs_path($monitoryaml);
    $ENV{'ARKUTIL_SPEC_PATH'} = $monitoryaml;

    $datadir = "$dataparent/run-$run_id";
    $andir = "$anparent/run-$run_id";
    $monitortxt = "$andir/monitors.txt";

    -d $datadir || mkdir $datadir;
    -d $andir || mkdir $andir;

    open($STATS, '>>', "$andir/stats.txt");
    print $STATS "# command line: $cmdline\n";

    $SIG{__DIE__} = sub {
	print STDERR "$Script died in step $step.\n";
    };

    if ($distrib) {
	open(OUT, '>', $monitortxt);
	my $cmd = "$bindir/manage-midar --list-monitors --activity=midar";
	if ($exclude_monitors) {
	    $cmd .= " --exclude=$exclude_monitors";
	}
	open(IN, '-|', $cmd) or die "manage-midar: $!\n";
	while (<IN>) {
	    if (/^(\S+)\s+([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)\s*$/) {
		push @monitors, $1;
		print OUT $_;
	    }
	}
	die "No monitors found.\n" unless (@monitors);
	$MIDAR_MONS=join(',', @monitors);
	print STDERR "ping\n";
	close IN;
	close OUT;
    }
}


sub midar_step() {
    if ($step eq 'remote-start') {
	start_step();
	if ($distrib) {
	    remote_manage("--kill-tmux");
	    remote_manage("--setup-tmux");
	    remote_manage("--start-marinda --start-shell-server");
	    remote_manage("--start-mper --pps=$mper_pps" .
		($simulate ? " --simulate" : ""));
	}

	next_step('est-targets');

    } elsif ($step eq 'est-targets') {
	if ((scalar @targetfiles < 1) && !-f "$andir/target-digest/list") {
	    die "Error: no targets specified.\n";
	}
	start_step("$andir/estimation-targets");
	runprog("$bindir/midar-targets",
	    'file', "$stepdir/targets.err",
	    "/usr/bin/time $bindir/midar-targets " .
	    "-s " .
	    ($distrib ? "-M $monitortxt " : "") .
	    ($bogons ? "-B $bogons " : "") .
	    "-A " . join(' ', @targetfiles) . " " .
	    (-f "$andir/target-digest/list" ? "-S @$andir/target-digest/list " : "") .
	    "-O $stepdir/targets");

	my $n_targets = `cat $stepdir/targets.*.targets | egrep -c '^[^#]'`;
	die "No targets found.\n" unless $n_targets;
	chomp $n_targets;
	print $STATS "targets: $n_targets\n";

	next_step('est-probe');

    } elsif ($step eq 'est-probe') {
	start_step("$datadir/estimation-data");
	my $num_methods = (-f "$andir/target-digest/list" ? 4 : 3);
	my $window_size = floor($est_round_duration * $mper_pps / $num_methods);
	my $n_targets = `cat $andir/estimation-targets/targets.*.targets | egrep -c '^[^#]'`;
	if ($distrib) {
	    $n_targets = ceil($n_targets / scalar @monitors);
	}
	my $n_windows = ceil($n_targets / $window_size);
	my $timeout = 0.9;
	my $eta = $n_windows * ($est_round_duration + $timeout) * $est_rounds;
	print STDERR "Estimated probing time: ${eta}s\n";

	if ($distrib) {
	    remote_manage("--preprocess-est-targets");
	    remote_manage("--upload-est-targets");

	    remote_probe(
		"--start-est-prober " .
		"--est-round-window=$window_size " .
		"--est-rounds=$est_rounds " .
		"--est-round-duration=$est_round_duration");

	    remote_manage("--download-est-results");
	    remote_manage("--delete-est-all");

	} else {
	    runprog("$bindir/midar-estimation-prober",
		'file', "$stepdir/local.err",
		"$bindir/midar-estimation-prober " .
		"--mper-port=$mper_port " .
		"--targets=$andir/estimation-targets/targets.0.0.0.0.targets " .
		"--round-window=$window_size " .
		"--rounds=$est_rounds " .
		"--round-duration=$est_round_duration " .
		"--log=$stepdir/local.log " .
		">$stepdir/local.out");
	}

	next_step('est-an');

    } elsif ($step eq 'est-an') {
	# estimation analysis

	# # analysis is faster when files are uncompressed
	# execute mkdir -p $tmpdir/estimation-data
	# andir=$tmpdir/estimation-data
	# files=""
	# for m in $(monlist); do
	#     files="$files $datadir/estimation-data/$m.estimation.$date.out.gz"
	# done
	# bg_wait_gzfiles "$andir" $files
	# reap "waiting for gzipped files" || exit $?

	# for f in $files; do
	#     outfile=$andir/$(basename $f .gz)
	#     if egrep -q "^## run finished" "$outfile"; then continue; fi
	#     echo "end marker not found in $outfile" >&2
	#     exit 1
	# done

	start_step("$andir/estimation-analysis");

	runprog("$bindir/estimation-analyze",
	    'file', "$stepdir/err",
	    "/usr/bin/time $bindir/estimation-analyze " .
	    ($distrib ? "--monitors $monitortxt " : "") .
	    "--gpscript " .
	    "--htmldir $stepdir " .
	    "--overlap $est_overlap " .
	    "$datadir/estimation-data/*.out*");

	analysis_plots($andir, "estimation-analysis");

	my $usable = `egrep -c '[*]' $stepdir/target-summary.txt`;
	print $STATS "usable targets: $usable\n";
	print $STATS "usable pairs: ", choose2($usable), "\n";

	next_step('disc-targets');

    } elsif ($step eq 'disc-targets') {

	start_step("$andir/discovery-targets");

	# distribute targets amongst monitors so that each monitor has the same velocity distribution
	runprog("$bindir/midar-targets",
	    'file', "$stepdir/targets.err",
	    "/usr/bin/time $bindir/midar-targets " .
	    "-u " .
	    ($distrib ? "-M $monitortxt " : "") .
	    (-f "$andir/target-digest/list" ? "-S @$andir/target-digest/list " : "") .
	    "-T $andir/estimation-targets/*.targets " .
	    "-E $andir/estimation-analysis/target-summary.txt " .
	    "-O $stepdir/targets");

	# generate sliding window schedule for discovery run
	runprog("$bindir/sliding-window.pl",
	    'file', "$stepdir/window.err",
	    "$bindir/sliding-window.pl " .
	    "--rate $mper_pps " .
	    "--gpscript " .
	    "--htmldir $andir/html/discovery-targets " .
	    "$stepdir/*.targets");

	print STDERR "plotting report graphs...\n";
	system("gnuplot $andir/html/discovery-targets/script.gp");
	print "Reports written to $andir/html/discovery-targets\n\n";

	my $n_targets = `cat $stepdir/targets.*.targets | egrep -c '^[^#]'`;
	chomp $n_targets;
	print $STATS "targets: $n_targets\n";
	print $STATS `egrep '^[^#]' $stepdir/window.err`;

	next_step('disc-probe');

    } elsif ($step eq 'disc-probe') {

	start_step("$datadir/discovery-data");

	print STDERR "Scheduled probing ",
	    `egrep '^time:' $andir/discovery-targets/window.err`;

	if ($distrib) {
	    remote_manage("--preprocess-disc-targets");
	    remote_manage("--upload-disc-targets");
	    remote_probe("--start-disc-prober");
	    remote_manage("--download-disc-results");
	    remote_manage("--delete-disc-all");

	} else {
	    runprog("$bindir/midar-discovery-prober",
		'file', "$stepdir/local.err",
		"$bindir/midar-discovery-prober " .
		"--schedule=$andir/discovery-targets/targets.0.0.0.0.targets.window " .
		"--targets=$andir/discovery-targets/targets.0.0.0.0.targets " .
		"--log=$stepdir/local.log " .
		"--mper-port=$mper_port " .
		"--no-sync-start " .
		">$stepdir/local.out");
	}

	next_step('disc-an');

    } elsif ($step eq 'disc-an') {
	# discovery analysis

	start_step("$andir/discovery-analysis");

	runprog("$bindir/resolution-analyze",
	    'tee', "$stepdir/err",
	    "$nice /usr/bin/time $bindir/resolution-analyze " .
	    ($distrib ? "--monitors $monitortxt " : "") .
	    "--gpscript " .
	    "--htmldir $stepdir " .
	    "--out $stepdir/comparisons " .
	    "--threads $n_threads " .
	    "--overlap $disc_overlap " .
	    "--responsive $disc_responsive " .
	    "--clockoff $clockoff " .
	    "$datadir/discovery-data/*.out*");

	analysis_plots($andir, "discovery-analysis");

	my $n_shared = `egrep -c '^[^#].* SHARED' $stepdir/comparisons`;
	chomp $n_shared;
	print $STATS "shared pairs: $n_shared\n";
	if ($n_shared == 0) {
	    print "No shared-counter pairs found.  Nothing left to do.\n\n";
	    last;
	}

	# extract shared pairs
	runprog("$bindir/transitive-closure",
	    'file', "$stepdir/pairs.err",
	    "/usr/bin/time $bindir/transitive-closure " .
	    "-f " .
	    "-o $stepdir/pairs " .
	    "$stepdir/comparisons");

	next_step('elim-targets', '1');

#    # transitive closure of shared pairs in disc-an
#    execute /usr/bin/time $bindir/transitive-closure \
#	-o $stepdir/tc \
#	$stepdir/comparisons \
#	2>$stepdir/tc.err


#disc-an-preview)
#    # discovery analysis preview
#
#    # make sure disc-an is done initial analysis
#    while true; do
#	if egrep -q '^# alias resolution' $stepdir/err; then break; fi
#	sleep 60
#    done
#
#    # plot graphs (single threaded, to not steal too many cycles from disc-an)
#    (
#	cd $stepdir
#	mkdir -p gp
#	for f in *.gp; do gnuplot $f; mv $f gp; done
#	execute mkdir html
#	execute mv *.html *.png html
#    )
#    ;;


    } elsif ($step eq 'elim-targets') {

	check_run_num();

	start_step("$andir/elim${num}-targets");

	my $pairfile;

	if ($num == 1) {
	    # first run: use pairs from discovery
	    $pairfile="$andir/discovery-analysis/pairs";
	} else {
	    # otherwise, use UNTESTED pairs from previous elim
	    $pairfile = "$andir/elim" . ($num-1) . "-analysis/untested-pairs";
	    # Previous elim would give a more recent velocity, but some
	    # targets may not have a velocity in previous elim, so we still
	    # need to use estimation velocity.
	}

	runprog("$bindir/generate-subsets",
	    'file', "$stepdir/subsets.err",
	    "$bindir/generate-subsets " .
	    "-o $stepdir/subsets " .
	    "-v $andir/estimation-analysis/target-summary.txt " .
	    "$pairfile");

	runprog("$bindir/midar-cor-targets.pl",
	    'file', "$stepdir/err",
	    "/usr/bin/time $bindir/midar-cor-targets.pl " .
	    "--out $stepdir " .
	    ($distrib ? "-M $monitortxt " : "") .
	    "--separate " .
	    "--shuffle " .
	    "$stepdir/subsets " .
	    "$andir/discovery-targets/*.targets");

	my $n_targets = `cat $stepdir/targets.*-src | egrep -c '^[^#]'`;
	chomp $n_targets;
	print $STATS "targets (with repeats): $n_targets\n";

	next_step('elim-probe', $num);

    } elsif ($step eq 'elim-probe') {

	check_run_num();
	start_step("$datadir/elim${num}-data");
	cor_style_probe('elim', $elim_rounds, $elim_concurrency);
	next_step('elim-an', $num);

    } elsif ($step eq 'elim-an') {

	check_run_num();

#    datafiles=""
#    echo "waiting for data files to finish..."
#    for m in $(monlist); do
#	f=$datadir/elim$num-data/$m.corroboration.one-src.$date-$num.out.gz
#	wait_file $f
#	datafiles="$datafiles $f"
#    done
#    for m in $(monlist); do
#	f=$datadir/elim$num-data/$m.corroboration.multi-src.$date-$num.out.gz
#	wait_file $f
#	datafiles="$datafiles $f"
#    done
#    echo "data files finished"

	start_step("$andir/elim${num}-analysis");

	my ($pairfile);

	if ($num == 1) {
	    # first run: use pairs from discovery
	    $pairfile="$andir/discovery-analysis/pairs";
	} else {
	    # otherwise, use pairs from previous elim
	    $pairfile="$andir/elim" . ($num-1) . "-analysis/pairs";
	}

	my @datafiles = (bsd_glob("$datadir/elim${num}-data/*one-src*out*"),
	    bsd_glob("$datadir/elim${num}-data/*multi-src*out*"));

	runprog("$bindir/resolution-analyze",
	    'tee', "$stepdir/err",
	    "$nice /usr/bin/time $bindir/resolution-analyze " .
	    ($distrib ? "--monitors $monitortxt " : "") .
	    "--cmp-nonoverlap " .
	    "--gpscript " .
	    "--html $stepdir " .
	    "--out $stepdir/comparisons " .
	    "--onlypairs $pairfile " .
	    "--threads $n_threads " .
	    "--overlap $elim_overlap " .
	    "--clockoff $clockoff " .
	    "--write-all " .
	    join(' ', @datafiles));

	analysis_plots($andir, "elim${num}-analysis");

	# transitive closure of shared pairs in disc-an and elim*-an
	runprog("$bindir/transitive-closure",
	    'file', "$stepdir/tc.err",
	    "/usr/bin/time $bindir/transitive-closure " .
	    "-o $stepdir/tc " .
	    "$andir/discovery-analysis/comparisons " .
	    "$andir/elim[1-${num}]-analysis/comparisons");

	# extract shared pairs
	runprog("$bindir/transitive-closure",
	    'file', "$stepdir/pairs.err",
	    "/usr/bin/time $bindir/transitive-closure " .
	    "-f " .
	    "-o $stepdir/pairs " .
	    "$andir/discovery-analysis/comparisons " .
	    "$andir/elim[1-${num}]-analysis/comparisons");

	runprog("$bindir/find-untested-pairs.pl",
	    'file', "$stepdir/untested-pairs.err",
	    "$bindir/find-untested-pairs.pl " .
	    "--include $andir/elim${num}-targets/subsets " .
	    "$andir/discovery-analysis/pairs " .
	    "$andir/elim[1-${num}]-analysis/comparisons " .
	    ">$stepdir/untested-pairs");

	my $n_shared = `egrep -c '^[^#]' $stepdir/pairs`;
	chomp $n_shared;
	print $STATS "shared pairs: $n_shared\n";

	# If many more pairs can still be eliminated, do "elim-targets $num+1".
	# Otherwise, move on to "elim-final".

	next_step('elim-final');

    } elsif ($step eq 'elim-final') {
	start_step("$andir/elim-final");
	my $elim_an_files = join ' ', bsd_glob("$andir/elim?-analysis/comparisons");

	if ($keep_small_untested) {
	    # Eliminate failed pairs, and untested pairs that belong to large sets
	    runprog("$bindir/transitive-closure",
		'file', "$stepdir/disc-elim.tc.err",
		"$bindir/transitive-closure " .
		"-s50 " .
		"-o$stepdir/disc-elim.tc " .
		"$andir/discovery-analysis/comparisons " .
		$elim_an_files);

	    runprog("$bindir/find-untested-pairs.pl",
		'file', "$stepdir/untested.small.pairs.err",
		"$bindir/find-untested-pairs.pl " .
		"--include $stepdir/disc-elim.tc-small " .
		"$andir/discovery-analysis/pairs " .
		"$elim_an_files " .
		">$stepdir/untested.small.pairs");

	    runprog("$bindir/transitive-closure",
		'file', "$stepdir/pairs.err",
		"$bindir/transitive-closure " .
		"-f " .
		"-o$stepdir/pairs " .
		"$stepdir/untested.small.pairs " .
		"$elim_an_files");

	    runprog("$bindir/transitive-closure",
		'file', "$stepdir/tc.err",
		"$bindir/transitive-closure " .
		"-o$stepdir/tc " .
		"$stepdir/untested.small.pairs " .
		"$elim_an_files");

	} else {
	    # Eliminate failed pairs and untested pairs
	    runprog("$bindir/transitive-closure",
		'file', "$stepdir/pairs.err",
		"$bindir/transitive-closure " .
		"-f " .
		"-o$stepdir/pairs " .
		$elim_an_files);

	    runprog("$bindir/transitive-closure",
		'file', "$stepdir/tc.err",
		"$bindir/transitive-closure " .
		"-o$stepdir/tc " .
		$elim_an_files);
	}

	my $n_pairs = `egrep -c '^[^#]' $stepdir/pairs`;
	chomp $n_pairs;
	print $STATS "surviving pairs: $n_pairs\n";

	next_step('cor-targets', '1');

    } elsif ($step eq 'cor-targets') {

	check_run_num();

	# Generate target lists for corroboration run.
	start_step("$andir/cor${num}-targets");

	my $prevdir = "$andir/elim-final";

#	if ($num > 1) {
#	    # use pairs from previous cor
#	    $prevdir="$andir/cor" . ($num-1) . "-analysis";
#	}

	runprog("$utildir/sets-to-pairs",
	    'tee', "$stepdir/tc.pairs.err",
	    "$utildir/sets-to-pairs " .
	    "$prevdir/tc " .
	    ">$stepdir/tc.pairs");

	# Note: estimation is the only step guaranteed to have a velocity for
	# every target.  Accuracy of velocity is not critical here; it's only
	# used to limit the subset size.
	runprog("$bindir/generate-subsets",
	    'file', "$stepdir/subsets.err",
	    "$bindir/generate-subsets " .
	    "-o $stepdir/subsets " .
	    "-v $andir/estimation-analysis/target-summary.txt " .
	    "$stepdir/tc.pairs");

	runprog("$bindir/midar-cor-targets.pl",
	    'file', "$stepdir/err",
	    "/usr/bin/time $bindir/midar-cor-targets.pl " .
	    "--out $stepdir " .
	    ($distrib ? "-M $monitortxt " : "") .
	    "--separate " .
	    "--shuffle " .
	    "$stepdir/subsets " .
	    "$andir/discovery-targets/*.targets");

	my $n_targets = `cat $stepdir/targets.*-src | egrep -c '^[^#]'`;
	chomp $n_targets;
	print $STATS "targets (with repeats): $n_targets\n";

	next_step('cor-probe', $num);

    } elsif ($step eq 'cor-probe') {

	check_run_num();
	start_step("$datadir/cor${num}-data");
	cor_style_probe('cor', $cor_rounds, $cor_concurrency);
	next_step('cor-an', $num);

    } elsif ($step eq 'cor-an') {

	check_run_num();

##    echo "waiting for data files to finish..."
##    for m in $(monlist); do
##	wait_file $datadir/cor$num-data/$m.corroboration.one-src.$date-$num.out.gz
##    done
##    for m in $(monlist); do
##	wait_file $datadir/cor$num-data/$m.corroboration.multi-src.$date-$num.out.gz
##    done
##    echo "data files finished"

	start_step("$andir/cor${num}-analysis");

	my @datafiles = (bsd_glob("$datadir/cor${num}-data/*one-src*out*"),
	    bsd_glob("$datadir/cor${num}-data/*multi-src*out*"));

	runprog("$bindir/resolution-analyze",
	    'tee', "$stepdir/err",
	    "/usr/bin/time $bindir/resolution-analyze " .
	    ($distrib ? "--monitors $monitortxt " : "") .
	    "--cmp-nonoverlap " .
	    "--gpscript " .
	    "--html $stepdir " .
	    "--out $stepdir/comparisons " .
	    "--onlypairs $andir/cor${num}-targets/tc.pairs " .
	    "--overlap $cor_overlap " .
	    "--clockoff $clockoff " .
	    "--write-all " .
	    "--alltest " .
	    join(' ', @datafiles));

	analysis_plots($andir, "cor${num}-analysis");

	my @elim_an_files = bsd_glob("$andir/elim?-analysis/comparisons");
	my @cor_an_files = map {"$andir/cor${_}-analysis/comparisons"} (1..$num);

	# find tc of elim and cor runs
	runprog("$bindir/transitive-closure",
	    'file', "$stepdir/tc.err",
	    "/usr/bin/time $bindir/transitive-closure " .
	    "-o $stepdir/tc " .
	    join(' ', @elim_an_files, @cor_an_files));

	next_step('midar-an');

    } elsif ($step eq 'midar-an') {

	start_step("$andir/midar-analysis");

	# find tc of pairs that survived all runs
	# Note: this is NOT just the tc of the cor run, because that might include
	# pairs that were introduced by tc when generating the cor targets from
	# disc and elim, but were never actually tested in disc or elim.  We want
	# only pairs that have passed disc/elim AND cor.
	
	my $cor_an_files = join ' ', bsd_glob("$andir/cor?-analysis/comparisons");

	runprog("$bindir/transitive-closure",
	    'file', "$stepdir/tc.err",
	    "/usr/bin/time $bindir/transitive-closure " .
	    "-P $andir/elim-final/pairs " .
	    "-o $stepdir/tc " .
	    "$cor_an_files");

	runprog("$bindir/transitive-closure",
	    'file', "$stepdir/pairs.err",
	    "/usr/bin/time $bindir/transitive-closure " .
	    "-f " .
	    "-P $andir/elim-final/pairs " .
	    "-o $stepdir/pairs " .
	    "$cor_an_files");

	my $n_pairs = `egrep -c '^[^#]' $stepdir/pairs`;
	chomp $n_pairs;
	print $STATS "pairs that survived Elim and passed Cor: $n_pairs\n";

	my @cor_data_dirs = bsd_glob("$datadir/cor?-data");
	my $cor_data_files = join(' \; ',
	    map { join(' ', map {chomp; $_} `ls -1tr $_/*one-src*.out*; ls -1tr $_/*multi-src*.out*`) } @cor_data_dirs);

	my @elim_data_dirs = bsd_glob("$datadir/elim?-data");
	my $elim_data_files = join(' \; ',
	    map { join(' ', map {chomp; $_} `ls -1tr $_/*one-src*.out*; ls -1tr $_/*multi-src*.out*`) } @elim_data_dirs);

	# check for transitive closure conflicts in all runs
	runprog("$bindir/set-analysis",
	    'file', "$stepdir/midar.sets.err",
	    "/usr/bin/time $bindir/set-analysis " .
	    "--output $stepdir/midar.sets " .
	    "--sets $stepdir/tc " .
	    "--overlap $cor_overlap " .
	    "--clockoff $clockoff " .
	    #"--graph $stepdir/midar-conflict " .
	    "$cor_data_files \\; " .
	    "$datadir/discovery-data/*.out* \\; " .
	    $elim_data_files);

	# remove conflicts
	runprog("$bindir/rm-conflicts.pl",
	    'tee', "$stepdir/midar-noconflicts.sets.err",
	    "/usr/bin/time $bindir/rm-conflicts.pl " .
	    "<$stepdir/midar.sets " .
	    ">$stepdir/midar-noconflicts.sets");

	# convert midar sets back to pairs
	runprog("$utildir/sets-to-pairs",
	    'tee', "$stepdir/midar-noconflicts.pairs.err",
	    "/usr/bin/time $utildir/sets-to-pairs " .
	    "$stepdir/midar-noconflicts.sets " .
	    ">$stepdir/midar-noconflicts.pairs");

	my $setsfile = "$stepdir/midar-noconflicts.sets";
	open SETS, "<", $setsfile;
	my $n_sets = 0;
	my $n_addrs = 0;
	while (<SETS>) {
	    if (/^# set/) {
		$n_sets++;
	    } elsif (/^[^#]/) {
		$n_addrs++;
	    }
	}
	close SETS;

	next_step($distrib ? 'remote-stop' : undef);

	print "#########################################################\n";
	print "MIDAR alias resolution complete.\n";
	print "Found $n_sets alias sets containing $n_addrs addresses.\n";
	print "Results are in $setsfile.\n";
	print $STATS "Alias sets: $n_sets\n";
	print $STATS "Addresses in alias sets: $n_addrs\n";

    } elsif ($step eq 'remote-stop') {
	start_step();
	if ($distrib) {
	    remote_manage("--kill-marinda --kill-shell-server");
	    remote_manage("--kill-mper");
	    remote_manage("--kill-tmux");
	}

	next_step(undef);

    } else {
	return 0;
    }

    return 1;
}

1;
